#! /usr/local/Gambit-C/bin/gsi-script

; File: "web-server.scm", Time-stamp: <2005-01-04 09:46:10 feeley>

; Copyright (C) 2004-2005 by Marc Feeley, All Rights Reserved.

; A minimal web server making it possible to access the filesystem
; using a browser.

(define (main arg)
  (let* ((port-number
          (string->number arg))
         (connection-queue
          (open-tcp-server
           (list port-number: port-number
                 reuse-address: #t
                 eol-encoding: 'cr-lf))))
    (let loop ()
      (let ((connection (read connection-queue)))
        (thread-start! (make-thread (lambda () (serve connection)))))
      (loop))))

(define (serve connection)
  (let* ((request (read-line connection))
         (x (parse request)))
    (let ((len (length x)))
      (if (and (>= len 2)
               (string=? (car x) "GET"))
        (display (get (cadr x)) connection)
        (display (unknown-request request) connection)))
    (close-port connection)))

(define (parse str)
  (call-with-input-string str (lambda (p) (read-fields p #\space #f))))

(define (read-fields port sep inc-sep?)
  (read-all port (lambda (p) (read-line p sep inc-sep?))))

(define (unknown-request request)
  (<html>
   "The following request was received: "
   (<code> request)))

(define (get path)
  (with-exception-catcher
   (lambda (exc)
     (<html>
      (<pre>
       (call-with-output-string
        ""
        (lambda (port) (##display-exception exc port))))))
   (lambda ()
     (get-path path))))

(define (get-path path)
  (let ((type (file-type path)))
    (case type
      ((directory)
       (<html>
        (<ol>
         (map (lambda (fn)
                (<li>
                 (<a> (path-expand fn path) fn)
                 "("
                 (object->string
                  (with-exception-catcher
                   (lambda (exc) 'other)
                   (lambda () (file-type (path-expand fn path)))))
                 ")"))
              (directory-files
               (list path: path ignore-hidden: #f))))))
      ((regular)
       (let* ((port (open-input-file path))
              (file (read-line port #f)))
         (close-input-port port)
         (let ((ext (path-extension path)))
           (if (or (string-ci=? ext ".htm")
                   (string-ci=? ext ".html"))
               file
               (<html> (<pre> file))))))
      (else
       (<html>
        "Only directories and regular files can be displayed")))))

(define (<html> . content)
  (list "<html>\n" (map escape content) "</html>\n"))

(define (<a> url . content)
  (list "<a href=\"" url "\">" (map escape content) "</a>\n"))

(define (<ol> . content)
  (list "<ol>\n" (map escape content) "</ol>\n"))

(define (<li> . content)
  (list "<li>\n" (map escape content) "\n"))

(define (<pre> . content)
  (list "<pre>\n" (map escape content) "</pre>\n"))

(define (<code> . content)
  (list "<code>" (map escape content) "</code>"))

(define (escape obj)
  (if (string? obj)
      (map esc (string->list obj))
      obj))

(define (esc ch)
  (case ch
    ((#\<) "&lt;")
    ((#\>) "&gt;")
    ((#\&) "&amp;")
    (else  ch)))
